(*
 * Copyright (c) 2003 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	Wroclaw and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *)
open Ast
open Util
open Pp

let xf = Printf.sprintf
let pf = Printf.printf

let gen_var =
  let cnt = ref 1 in
  fun () -> 
    assert (!cnt < 1000000000); incr cnt; !cnt,ref BNone

let mr n = {node=n;clone=None}

let rf x = mr (Var x)

let var_id (i,_) = i



let rec var_ids = function
 | [] -> []
 | v::vs -> (var_id v)::(var_ids vs)

let rec lambda x y = 
  match x with
  | [] -> y
  | x :: xs -> mr (Abs (x, lambda xs y))

let rec app = function
  | [] -> assert false
  | [x] ->  x
  | x1 :: x2 :: xs -> app ( mr (App (x1, x2)) :: xs)
  
let var_vect lst = List.map (fun x -> gen_var (), x) lst


let rec clear = function
  | [] -> ()
  | (x::xs) -> clear xs; x.clone<-None


let freshen t=
  let clearlist = ref [] in
  let rec f map l =
    match l with
    | {clone=Some c} -> c
    | {node=App (t1,t2)} -> 
      let c=mr (App (f map t1, f map t2)) in clearlist:=l::!clearlist; l.clone<-Some c; c
    | {node=Abs (k, t)} ->
      let k' = gen_var () in 
      let c=mr (Abs (k', f (Intmap.add (var_id k) k' map) t)) in clearlist:=l::!clearlist; l.clone<-Some c; c
    | {node=Var n} when Intmap.mem (var_id n) map -> let c=mr (Var (Intmap.find (var_id n) map)) in 
         clearlist:=l::!clearlist; l.clone<-Some c; c
    | t -> t in 
   let nt=f Intmap.empty t in clear !clearlist; clearlist:= []; nt
    
      
let var_value (_,b) =
  match !b with
  | BCached t -> t
  | BSusp f -> let t= f() in b:=BCached t;t
  | BNone -> failwith "Unbound value"

let rec print_term_list = function
   | [] -> ()
   | (x::xs) -> print_string ("T: "^(string_of_term x)^"\n"); print_term_list xs
 
let id x = x

let op2 o a1 a2=
 if o="+" then a2+a1 else
 if o="-" then a2-a1 else
 if o="*" then a2*a1 else
 if o="/" then a2/a1 else
   assert false
let dummy_value = {val_name = "???"; val_ty = Tyvar "'?"; val_semantics = Dummy}


let rec whnf t =
  let rec f ap t =
    match t with
    | {node=Var ((_,b) as x)} -> 
    begin
      match !b with
      | BNone -> app (t::ap)
      | _ -> f ap (var_value x)
    end
    | {node=App(e1,e2)} -> f (e2::ap) e1
    | {node=Abs(v,e)} ->
      begin
        match ap with
        | [] -> t
        | a::ass ->
          begin
            match freshen t with
             | {node=Abs((_,nb),ne)} -> nb:=BSusp (function () -> whnf a); f ass ne
             | _ -> failwith "Internal error"
          end
      end
    | {node=Value ({val_semantics = Eliminator tc; val_name = n})} when n="*"||n="/"||n="+"||n="-" ->
    begin
      match List.map whnf ap with
      | {node=Value {val_semantics=Integer a1}}::{node=Value {val_semantics=Integer a2}}::ass -> 
        let na=op2 n a1 a2 in 
           app ((mr (Value {dummy_value with val_semantics= Integer na} ))::ass)
      | x -> app(t::x)
    end
    | {node=Value ({val_semantics = Eliminator tc} as ev)} ->
    begin
      match ap with
      | [] -> t
      | a::ass ->
        let rec find_ctor acc = function
          | {node=App (e1, e2)} -> find_ctor (e2 :: acc) e1
          | {node=Value ({val_semantics = (Integer v)})} when tc.tc_name="NAT"->
            if (v>0) then              
              let our_rule r = r.r_elim == ev && r.r_ctor.val_name = "@S"
              in Some (List.find our_rule tc.tc_rules, (mr (Value {dummy_value with val_semantics= Integer (v-1)} ))::acc)
              else
              let our_rule r = r.r_elim == ev && r.r_ctor.val_name = "@O"
              in Some (List.find our_rule tc.tc_rules, acc)
            
          
          | {node= Value ({val_semantics = (Constructor tc')} as cv)} when tc == tc' ->
            let our_rule r = r.r_elim == ev && r.r_ctor == cv in
            Some (List.find our_rule tc.tc_rules, acc)
          | _ -> None
        in
        begin
          match find_ctor [] (whnf a) with
          | (Some (r, args)) when List.length args == List.length r.r_vars ->
              f ass (f args (lambda r.r_vars r.r_value)) 
          | _ -> app (t::ap)
        end           
    end
    | {node=(Value _)} -> app (t::ap)
  in
    f [] t
  
let rec complete_development t =
  let rec f ap t =
    match t with
    | {node=Var ((_,b) as x)} -> 
    begin
      match !b with
      | BNone -> app (t::(List.map complete_development ap))
      | _ -> f ap (var_value x)
    end
    | {node=App(e1,e2)} -> f (e2::ap) e1
    | {node=Abs(v,e)} ->
      begin
        match ap with
        | [] -> 
        begin
          let is_fv t =
            let clearlist = ref [] in
            let rec is_fv2 l =
              match l with
              | {clone=Some x} -> false
              | {node=App (t1, t2)} -> if is_fv2 t1 || is_fv2 t2 then true else begin clearlist:=l::!clearlist; l.clone<-Some l;false end
              | {node=Abs (n, t)} -> if n != v && is_fv2 t then true else begin clearlist:=l::!clearlist; l.clone<-Some l;false end
              
              | {node=Var n} -> if n==v then true else begin clearlist:=l::!clearlist; l.clone<-Some l;false end
              | {node=Value _} -> false
            in
              let r=is_fv2 t in clear !clearlist; clearlist:= []; r
          in  
          match (complete_development e) with
          | {node=App(e1,{node=Var v'})} when v==v' && not (is_fv e1) -> e1
          | e' -> lambda [v] e'
        end
        | a::ass ->
          begin
            match freshen t with
             | {node=Abs((_,nb),ne)} -> nb:=BSusp (function () -> complete_development a); f ass ne
             | _ -> failwith "Internal error"
          end
      end
    | {node=Value ({val_semantics = Eliminator tc; val_name = n})} when n="*"||n="/"||n="+"||n="-" ->
    begin
      match List.map complete_development ap with
      | {node=Value {val_semantics=Integer a1}}::{node=Value {val_semantics=Integer a2}}::ass -> 
        let na=op2 n a1 a2 in 
           app ((mr (Value {dummy_value with val_semantics= Integer na} ))::ass)
      | x -> app(t::x)
    end
    | {node=Value ({val_semantics = Eliminator tc} as ev)} ->
    begin
      match ap with
      | [] -> t
      | a::ass ->
        let rec find_ctor acc = function
          | {node=App (e1, e2)} -> find_ctor (e2 :: acc) e1
          | {node=Value ({val_semantics = (Integer v)})} when tc.tc_name="NAT"->
            if (v>0) then              
              let our_rule r = r.r_elim == ev && r.r_ctor.val_name = "@S"
              in Some (List.find our_rule tc.tc_rules, (mr (Value {dummy_value with val_semantics= Integer (v-1)} ))::acc)
              else
              let our_rule r = r.r_elim == ev && r.r_ctor.val_name = "@O"
              in Some (List.find our_rule tc.tc_rules, acc)
            
          
          | {node= Value ({val_semantics = (Constructor tc')} as cv)} when tc == tc' ->
            let our_rule r = r.r_elim == ev && r.r_ctor == cv in
            Some (List.find our_rule tc.tc_rules, acc)
          | _ -> None
        in
        begin
          match find_ctor [] (whnf a) with
          | (Some (r, args)) when List.length args == List.length r.r_vars ->
              f ass (f args (lambda r.r_vars r.r_value)) 
          | _ -> app (t::(List.map complete_development ap))
        end
           
    end
    | {node=(Value _)} -> app (t::(List.map complete_development ap))
  in
    f [] (whnf t)


let nf t = complete_development t

(*let nf t = 
  let rec f t =
    print_string ("before:"^(string_of_term t)^"\n");
    let (didsth, t') = complete_development t in
    print_string ("after:"^(string_of_term t')^"\n");
    if didsth then f t' else t'
  in f t*)

exception Different

let eq t1 t2 =
  let rec f m = function
    | {node=App (t1, t2)}, {node=App (t1', t2')} ->
      f (f m (t1, t1')) (t2, t2')
    | {node=Abs (k, t)}, {node=Abs (k', t')} ->
      f (Intmap.add (var_id k') k m) (t, t')
    | {node=Var k}, {node=Var k'} when k == Intmap.find (var_id k') m -> m
    | {node=Value v}, {node=Value v'} when v == v' -> m
    | _ -> raise Different
  in
  try
    let _ = f Intmap.empty (t1, t2) in
    true
  with Different -> false
  
  
let make_rule elim con vars t =
  {r_vars = vars; 
   r_elim = elim; 
   r_ctor = con; 
   r_value = nf t}

